package acube.gram.backend.birt;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpSession;

import org.apache.log4j.Logger;
import org.eclipse.birt.report.engine.api.EngineException;
import org.eclipse.birt.report.engine.api.HTMLRenderOption;
import org.eclipse.birt.report.engine.api.IGetParameterDefinitionTask;
import org.eclipse.birt.report.engine.api.IParameterDefnBase;
import org.eclipse.birt.report.engine.api.IParameterGroupDefn;
import org.eclipse.birt.report.engine.api.IParameterSelectionChoice;
import org.eclipse.birt.report.engine.api.IReportEngine;
import org.eclipse.birt.report.engine.api.IReportRunnable;
import org.eclipse.birt.report.engine.api.IRunAndRenderTask;
import org.eclipse.birt.report.engine.api.IScalarParameterDefn;
import org.eclipse.birt.report.model.api.CascadingParameterGroupHandle;
import org.eclipse.birt.report.model.api.ReportDesignHandle;
import org.eclipse.birt.report.model.api.ScalarParameterHandle;

import acube.framework.vo.ErrorVO;
import acube.gram.backend.ExecutionContext;
import acube.gram.constant.DataType;
import acube.gram.constant.InputType;
import acube.gram.constant.ReportFormat;
import acube.gram.constant.ReportFormatCodec;
import acube.gram.exception.GramErrorCodes;
import acube.gram.exception.GramException;
import acube.gram.exception.GramExceptionFactory;
import acube.gram.vo.FileVO;
import acube.gram.vo.OptionVO;
import acube.gram.vo.ParameterVO;

/**
 * Implmentation pour BIRT
 * @author MAE
 * @see acube.gram.backend.ExecutionContext
 */
public class BirtExecutionContext implements ExecutionContext {

	protected Logger logger = Logger.getLogger(this.getClass());
	protected String THIS_CLASS = "["+this.getClass()+"] ";
	
	/**
	 * la session de la requte initiale
	 */
	private HttpSession session;

	/**
	 * instance de BIRT
	 */
	private IReportEngine engine;
	
	/**
	 * le rapport en cours d'excution
	 */
	private IReportRunnable design;
	
	/**
	 * Paramtres du rapport avec les valeurs dfinies par le client
	 */
	private Map<String,ParameterVO> parameters = new HashMap<String,ParameterVO>();
	

	/**
	 * Constructeur
	 * @param engine
	 * @param session
	 * @param runnable
	 * @throws GramException 
	 */
	public BirtExecutionContext(IReportEngine engine,HttpSession session,IReportRunnable runnable)
			throws GramException {

		this.engine = engine;
		this.session = session;
		this.design = runnable;
		
		IGetParameterDefinitionTask task = engine.createGetParameterDefinitionTask(design);
		Iterator params = task.getParameterDefns(true).iterator();
		while (params.hasNext()) {
			IParameterDefnBase param = (IParameterDefnBase)params.next();
			if (param instanceof IParameterGroupDefn)
				throw GramExceptionFactory.createUnimplementedOperation("parameter group");
			else {
				IScalarParameterDefn scalar = (IScalarParameterDefn) param;
				parameters.put(scalar.getName(),createParameterVO(task,scalar));
			}
		}
	}
	
	/**
	 * cre un ParameterVO pour GRaM  partir d'un IScalarParameterDefn de BIRT
	 * @param task 
	 * @param scalar
	 * @return
	 * @throws GramException
	 */
	private ParameterVO createParameterVO(IGetParameterDefinitionTask task, IScalarParameterDefn scalar)
			throws GramException {
		
		ParameterVO result = new ParameterVO();
		result.setName(scalar.getName());
		if (scalar.getPromptText() != null)
			result.setPrompt(scalar.getPromptText());
		else
			result.setPrompt(scalar.getName());
		result.setHidden(scalar.isHidden());

		// valeur par dfaut
		result.setCurrentValue(scalar.getDefaultValue());

		// type de donnes attendu
		result.setSecret(scalar.isValueConcealed());
		result.setBlankValueAllowed(scalar.allowBlank());
		result.setNullValueAllowed(scalar.allowNull());

		switch (scalar.getDataType()) {
		case IScalarParameterDefn.TYPE_STRING:
			result.setDataType(DataType.STRING);
			break;
		case IScalarParameterDefn.TYPE_FLOAT:
		case IScalarParameterDefn.TYPE_DECIMAL:
			result.setDataType(DataType.DOUBLE);
			break;
		case IScalarParameterDefn.TYPE_BOOLEAN:
			result.setDataType(DataType.BOOLEAN);
			break;
		case IScalarParameterDefn.TYPE_DATE_TIME:
			result.setDataType(DataType.DATETIME);
			break;
		default:
			result.setInputType(null);
			break;
		}
		
		// interface utilisateur
		switch (scalar.getControlType()) {
		case IScalarParameterDefn.TEXT_BOX:
			result.setInputType(InputType.TEXT);
			break;
		case IScalarParameterDefn.LIST_BOX:
			result.setInputType(InputType.LIST);
			break;
		case IScalarParameterDefn.RADIO_BUTTON:
			result.setInputType(InputType.RADIO);
			break;
		case IScalarParameterDefn.CHECK_BOX:
			result.setInputType(InputType.CHECK);
			break;
		default:
			result.setInputType(InputType.TEXT);
			break;
		}
		if(result.getInputType() !=  InputType.TEXT) {

			ReportDesignHandle reportHandle = (ReportDesignHandle)design.getDesignHandle( );
			ScalarParameterHandle parameterHandle = (ScalarParameterHandle)reportHandle.findParameter(scalar.getName());
			
			if (parameterHandle.getContainer( ) instanceof CascadingParameterGroupHandle){
				throw GramExceptionFactory.createUnimplementedOperation("cascading parameters");

			} else {
			   Collection selectionList = task.getSelectionList(scalar.getName());
			   if (selectionList != null) {
				   List<OptionVO> options = new ArrayList<OptionVO>();
				   Iterator i = selectionList.iterator();
				   while (i.hasNext()) {
					   IParameterSelectionChoice choice = (IParameterSelectionChoice) i.next();
					   OptionVO option = new OptionVO();
					   option.setLibelle(choice.getLabel());
					   option.setValue(choice.getValue().toString()); // TODO encoder proprement
					   options.add(option);
				   }
				   result.setPossibleValues(options);
			   }
			}
		}
		return result;
	}
	
	public Collection<ParameterVO> getParameters() throws GramException {

		return parameters.values();
	}
	
	public void setParameters(Map<String, String> params)
			throws GramException {
		
		for (Map.Entry<String, String> entry : params.entrySet()) {
			ParameterVO parameter = parameters.get(entry.getKey());
			if (parameter==null) {
				ErrorVO error = new ErrorVO();
				error.setCdError(GramErrorCodes.INVALID_PARAMETER);
				error.setParamLabel(entry.getKey());
				throw new GramException(error);
			}
			
			String value = entry.getValue().trim();
			parameter.setCurrentValue(value);

			List<OptionVO> possibleValues = parameter.getPossibleValues();
			// si c'est une liste droulante, on met les champs "selected"  jour
			// et on vrifie que la valeur est valide
			if (possibleValues != null) {
				boolean found = false;
				for (OptionVO option : possibleValues) {
					option.setSelected(option.getValue().equals(value));
					found |= option.isSelected();
				}
				parameter.setCurrentValueValid(found);
				
			// sinon, on vrifie seulement qu'elle n'est pas vide
			// TODO : vrifier le type de donne
			} else {
				parameter.setCurrentValueValid(parameter.isBlankValueAllowed() || !value.equals(""));
			}
		}
	}

	public FileVO execute(ReportFormat format) 
			throws GramException {
		
		IRunAndRenderTask task = engine.createRunAndRenderTask(design);

		for (Map.Entry<String,ParameterVO> entry : parameters.entrySet())
			task.setParameterValue(entry.getKey(),entry.getValue().getCurrentValue()); // TODO dcoder les chaines
		
		FileVO fileVO = new FileVO();
		HTMLRenderOption option = new HTMLRenderOption();

		switch (format) {
		case HTML_FRAGMENT:
			fileVO.setContentType("text/xml"); // TODO
			option.setOutputFormat(HTMLRenderOption.OUTPUT_FORMAT_HTML);
			option.setEmbeddable(true);
			break;
		case HTML:
			fileVO.setContentType("text/html"); // TODO
			option.setOutputFormat(HTMLRenderOption.OUTPUT_FORMAT_HTML);
			break;
		case PDF:
			fileVO.setContentType("application/pdf"); // TODO
			fileVO.setFileName("report.pdf");
			option.setOutputFormat(HTMLRenderOption.OUTPUT_FORMAT_PDF);
			break;
		default:
			{
				ErrorVO error = new ErrorVO();
				error.setCdError(GramErrorCodes.FORMAT_NOT_SUPPORTED);
				error.setParamLabel(ReportFormatCodec.encode(format));
				throw new GramException(error);
			}
		}
		
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		option.setOutputStream(os);
		task.setRenderOption(option);

		HashMap<String,Object> context = new HashMap<String, Object>();
		context.put(GramHTMLImageHandler.CONTEXT_KEY_HTTP_SESSION,session);
		task.setAppContext(context);
		
		try {
			task.run();
			
		} catch (EngineException e) {
			ErrorVO vo = new ErrorVO();
			vo.setCdError(GramErrorCodes.SERVICE_ERROR);
			vo.setParamLabel(e.getMessage());
			throw new GramException(vo);
		}

		fileVO.setContent(os.toByteArray());
		return fileVO;
	}

	/**
	 * non applicable  BIRT
	 */
	public void toggleItem(String itemID)
		throws GramException {
		
		throw new GramException(GramErrorCodes.UNIMPLEMENTED_OPERATION);
	}
}
