package acube.gram.backend.ssrs;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.xml.rpc.holders.ByteArrayHolder;
import javax.xml.rpc.holders.StringHolder;
import javax.xml.soap.SOAPException;

import org.apache.axis.message.SOAPHeaderElement;
import org.apache.log4j.Logger;

import acube.framework.exception.TechnicalException;
import acube.gram.backend.ExecutionContext;
import acube.gram.constant.DataType;
import acube.gram.constant.InputType;
import acube.gram.constant.ReportFormat;
import acube.gram.exception.GramErrorCodes;
import acube.gram.exception.GramException;
import acube.gram.exception.GramExceptionFactory;
import acube.gram.frontend.ActionHandler;
import acube.gram.frontend.ResourceHandler;
import acube.gram.vo.FileVO;
import acube.gram.vo.OptionVO;
import acube.gram.vo.ParameterVO;

import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ExecutionInfo;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ParameterStateEnum;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ParameterTypeEnum;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ParameterValue;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ReportExecutionServiceSoapStub;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ReportParameter;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.ValidValue;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.holders.ArrayOfStringHolder;
import com.microsoft.schemas.sqlserver._2005._06._30.reporting.reportingservices.holders.ArrayOfWarningHolder;

public class SSRSExecutionContext implements ExecutionContext {
	
	private final String THIS_CLASS = "[" + this.getClass().getName() + "]";
	static Logger logger = Logger.getLogger(SSRSWrapper.class);

	/**
	 * Namespace pour grer la session dans les appels au service web
	 */
	public static final String reNamespace = "http://schemas.microsoft.com/sqlserver/2005/06/30/reporting/reportingservices";

	/**
	 * le "pool" de connexions
	 */
	private SOAPStubPool pool;
	/**
	 * la session de la requte initiale
	 */
	private HttpSession session;
	/**
	 * l'identifiant d'excution dans Reporting Services
	 */
	private String executionID;
	private List<ParameterVO> parameterDefinitions;
	private Map<String,String> parameterValues = new HashMap<String, String>();;
	private ActionHandler actionHandler;
	private ResourceHandler resourceHandler;
	
	public SSRSExecutionContext(SOAPStubPool pool,
								HttpServletRequest request,
								ExecutionInfo info,
								ActionHandler actionHandler, 
								ResourceHandler resourceHandler) {
		
		this.pool = pool;
		this.session = request.getSession();
		this.executionID  = info.getExecutionID();
		this.actionHandler = actionHandler;
		this.resourceHandler = resourceHandler;
		
		setParameters(info);
	}

	private void setParameters(ExecutionInfo info) {
		
		// on efface la liste
		this.parameterDefinitions = new ArrayList<ParameterVO>();
		
		// on rcupre les paramtres  partir des dfinitions de Reporting Services
		for (ReportParameter rsParam : info.getParameters()) {
			if (rsParam.getPromptUser()) {
				this.parameterDefinitions.add(createParameterVO(rsParam));
			}
		}
	}

	private ParameterVO createParameterVO(ReportParameter rsParam) {
		
		ParameterVO result = new ParameterVO();
		
		result.setName(rsParam.getName());
		result.setPrompt(rsParam.getPrompt());
		
		// type de donne
		if (rsParam.getType() == ParameterTypeEnum.Boolean)
			result.setDataType(DataType.BOOLEAN);
		else if (rsParam.getType() == ParameterTypeEnum.String)
			result.setDataType(DataType.STRING);
		else if (rsParam.getType() == ParameterTypeEnum.DateTime)
			result.setDataType(DataType.DATETIME);
		else if (rsParam.getType() == ParameterTypeEnum.Integer)
			result.setDataType(DataType.INT);
		else if (rsParam.getType() == ParameterTypeEnum.Float)
			result.setDataType(DataType.DOUBLE);
		
		result.setNullValueAllowed(rsParam.getNullable());
		result.setBlankValueAllowed(rsParam.getAllowBlank());
		
		result.setCurrentValueValid(rsParam.getState()==ParameterStateEnum.HasValidValue);
		String currentValue = parameterValues.get(rsParam.getName());
		if (currentValue==null)
		{
			String[] defaultValues = rsParam.getDefaultValues();
			if ((defaultValues != null) && (defaultValues.length>0))
				currentValue = defaultValues[0];
		}
		result.setCurrentValue(currentValue);
		
		// Interface utilisateur
		ValidValue[] validValues = rsParam.getValidValues();
		if (validValues != null) {
			ArrayList<OptionVO> options = new ArrayList<OptionVO>();
			for (int j=0; j<validValues.length; j++) {
				String label = validValues[j].getLabel();
				String value = validValues[j].getValue();

				OptionVO optionVO = new OptionVO();
				optionVO.setLibelle(label);
				optionVO.setValue(value);
				optionVO.setSelected(value.equals(currentValue));

				options.add(optionVO);
			}
			result.setInputType(InputType.LIST);
			result.setPossibleValues(options);

		} else
			result.setInputType(InputType.TEXT);
		
		return result;
	}

	/**
	 * cre l'en-tte qui contient l'identifiant d'excution pour Reporting Services
	 * @param executionId
	 * @return l'lment d'en-tte
	 * @throws SOAPException
	 */
	public SOAPHeaderElement createSessionHeader(String executionId)
			throws SOAPException {
		
		SOAPHeaderElement sessionHeader = new SOAPHeaderElement(
				reNamespace, "ExecutionHeader");
		sessionHeader.addChildElement(new SOAPHeaderElement(reNamespace,
				"ExecutionID", executionId));
		return sessionHeader;
	}

	/**
	 * liste des paramtres
	 */
	public Collection<ParameterVO> getParameters() {

		return parameterDefinitions;
	}
	
	/**
	 * Dfinit les paramtres du rapport, pour pouvoir l'excuter
	 */
	public void setParameters(Map<String, String> params)
			throws GramException {

		logger.info(THIS_CLASS + "setParameters " + executionID);
		
		// on reformule les paramtres
		ArrayList<ParameterValue> parameters =
			new ArrayList<ParameterValue>();

		for (Map.Entry<String, String> entry : params.entrySet()) {
			
			String name = entry.getKey();
			String value = entry.getValue();
			
			logger.debug(THIS_CLASS + "param " + name + "=" + value);
			parameterValues.put(name, value);
			
			ParameterValue parameterValue = new ParameterValue();
			parameterValue.setName(name);
			parameterValue.setValue(value);
			parameters.add(parameterValue);
		}

		// on les met dans un tableau
		ParameterValue parameterArray[] = new ParameterValue[parameters.size()];
		parameterArray = parameters.toArray(parameterArray);

		try {
			// on appelle le service web
			ReportExecutionServiceSoapStub service = pool.getReportExecutionServiceSoapStub();
			// en rappelant 'identifiant d'excution
			service.setHeader(createSessionHeader(executionID));
			ExecutionInfo info = service.setExecutionParameters(parameterArray, "fr-fr");
			// et on met  jour la liste des paramtres
			setParameters(info);
			
		} catch (RemoteException e) {
			logger.error(THIS_CLASS + "RemoteException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);
			
		} catch (SOAPException e) {
			logger.error(THIS_CLASS + "SOAPException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);
		}
	}

	
	public FileVO execute(ReportFormat _format)
			throws GramException {

		logger.info(THIS_CLASS + "executeRapport " + executionID);
		try {
			// initialisation de l'appel au service web
			ReportExecutionServiceSoapStub service = pool.getReportExecutionServiceSoapStub();
			service.setHeader(createSessionHeader(executionID));
			
			// paramtres lis au format attendu
			String format;
			String deviceInfo = null;
			String fileName = null;
			
			switch (_format) {
			case HTML_FRAGMENT:
				// on ne peut pas mettre <HTMLFragment>True</HTMLFragment> car
				// les actions utilisateur sont dsactives. On fait donc une page HTML
				// complte, et il faudra enlever les niveaux HTML et BODY.
			case HTML:
				deviceInfo = "<DeviceInfo>";
				deviceInfo += "<Toolbar>False</Toolbar>";
				deviceInfo += "<Parameters>False</Parameters>";
				deviceInfo += "<Javascript>False</Javascript>";
				deviceInfo += "</DeviceInfo>";
				format = "HTML4.0";
				break;
				
			case PDF:
				format = "PDF";
				fileName = "rapport.pdf";
				break;
				
			case EXCEL:
				format = "EXCEL";
				fileName = "rapport.xls";
				break;
			
			default:
				throw new TechnicalException(GramErrorCodes.FORMAT_NOT_SUPPORTED);
			}
			
			// holders pour le rsultat
			ByteArrayHolder result = new ByteArrayHolder();
			StringHolder extensionH = new StringHolder();
			StringHolder mimeTypeH = new StringHolder();
			StringHolder encodingH = new StringHolder();
			ArrayOfWarningHolder warnings = new ArrayOfWarningHolder();
			ArrayOfStringHolder streamIds = new ArrayOfStringHolder();

			// rendu du rapport
			service.render(format, deviceInfo, result, extensionH, mimeTypeH,
					encodingH, warnings, streamIds);
			
			// rcupration des ressources externes (pour le HTML)
			// on les met dans un Map pour pouvoir les retrouver lors
			// de l'analyse des URLs du HTML produit
			Map<String,FileVO> streams = new HashMap<String, FileVO>();
			for (int i = 0; i < streamIds.value.length; i++) {
				String id = streamIds.value[i];
				FileVO vo = getStream(service,format,id,deviceInfo);
				streams.put(id, vo);
			}

			
			// Si le format est HTML, on remplace les URL
			byte[] content = result.value;
			String encoding = encodingH.value;
			switch (_format) {
			case HTML_FRAGMENT:
				{
					// TODO: l'encodage fourni par SSRS n'est pas standard
					encoding = "UTF-8";
					// dcodage
					String html = new String(result.value, encoding);
					// on retire les niveaux HTML et BODY
					html = extractBody(html);
					// on transforme les URLs
					content = transformUrls(html, streams).getBytes(encoding);
					break;
				}	
			case HTML:
				{
					// TODO: l'encodage fourni par SSRS n'est pas standard
					encoding = "UTF-8";
					// dcodage
					String html = new String(result.value, encoding);
					// on transforme les URLs
					content = transformUrls(html, streams).getBytes(encoding);
					break;
				}
			}

			// on met le rsultat dans un FileVO
			FileVO vo = new FileVO();
			//vo.setEncoding(encoding);
			vo.setContentType(mimeTypeH.value);
			vo.setFileName(fileName);
			vo.setContent(content);
			return vo;

		} catch (RemoteException e) {
			logger.error(THIS_CLASS + "RemoteException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);

		} catch (Exception e) {
			logger.error(THIS_CLASS + "Exception: " + e.getMessage());
			throw GramExceptionFactory.createUnknownError(e);
		}
	}

	/**
	 * rcupre une ressource externe au rapport (images, feuilles de style, scripts etc.)
	 * @param rs le stub du service web qui a servi a excuter le rapport (le header y est dfini)
	 * @param format le format demand pour le rapport (pas pour la ressource)
	 * @param streamId l'identifiant de la ressource
	 * @param deviceInfo les paramtres d'excution du rapport
	 * @return un FileVO contenant la ressource
	 * @throws TechnicalException
	 */
	public FileVO getStream(ReportExecutionServiceSoapStub rs, String format, String streamId, String deviceInfo)
			throws GramException {

		logger.info(THIS_CLASS+"get " + streamId);

		ByteArrayHolder result = new ByteArrayHolder();
		StringHolder encoding = new StringHolder();
		StringHolder mimeType = new StringHolder();

		try {
			rs.renderStream(format,
							streamId,
							deviceInfo,
							result,
							encoding,
							mimeType);
			FileVO vo = new FileVO();
			vo.setContentType(mimeType.value);
			vo.setContent(result.value);
			return vo;

		} catch (RemoteException e) {
			logger.error(THIS_CLASS + "RemoteException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);
		}
	}


	public void toggleItem(String itemID)
			throws GramException {

		logger.info(THIS_CLASS + "toggleItem " + itemID);
		try {
			ReportExecutionServiceSoapStub service = pool.getReportExecutionServiceSoapStub();
			service.setHeader(createSessionHeader(executionID));
			service.toggleItem(itemID);
		
		} catch (RemoteException e) {
			logger.error(THIS_CLASS + "RemoteException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);
		
		} catch (SOAPException e) {
			logger.error(THIS_CLASS + "SOAPException: " + e.getMessage());
			throw GramExceptionFactory.createServiceError(e);
		}
	}
	
	
	
	/*
	 * HTML transformation
	 */

	/**
	 * Retire les balises HTML, HEAD et BODY
	 * Remet les balises STYLE et le contenu de BODY dans un unique DIV
	 * @param html le HTML d'origine
	 * @return le fragment de HTML
	 */
	public String extractBody(String html) {

		StringBuffer buffer = new StringBuffer(html.length());
		buffer.append("<DIV style=\"overflow:auto;HEIGHT:100%;WIDTH:100%\">");
		
		// les feuilles de styles
		int styleStart = 0;
		while (true) {
			styleStart = html.indexOf("<style",styleStart);
			if (styleStart<0)
				break;
			int styleEnd = html.indexOf("</style>",styleStart)+8;
			buffer.append(html.substring(styleStart,styleEnd));
			styleStart = styleEnd;
		}

		// body
		int bodyStart,bodyEnd;
		bodyStart = html.indexOf("<body");
		bodyStart = html.indexOf(">",bodyStart)+1;
		bodyEnd = html.indexOf("</body>",bodyStart);
		buffer.append(html.substring(bodyStart,bodyEnd));

		buffer.append("</DIV>");
		return buffer.toString();
	}
	
	
	/**
	 * Remplace les URLs fournies par Reporting Services par des URL cres par le frontend
	 * @param html le html d'origine
	 * @param streams les ressources externes
	 * @return le HTML o les URLs ont t remplaces
	 */
	public String transformUrls(String html,
								Map<String,FileVO> streams)
			throws GramException {

		// On repre les URL qui font rfrence au serveur (pool.getUrl()),
		// et on isole la partie requte (aprs le ? et avant la fin de la chane)
		// TODO passer les autres URLs  actionHandler.handleHyperlink
		Pattern pattern = Pattern.compile(
				"("
				+ Pattern.quote(pool.getUrl() + "?")
				+ "[^\"]*)");
	
		StringBuffer buffer = new StringBuffer();
		int lastEnd = 0;
		Matcher matcher = pattern.matcher(html);
	
		// pour chaque URL
		while (matcher.find()) {
			int start = matcher.start();
			int end = matcher.end();
			// le texte avant l'URL
			buffer.append(html.substring(lastEnd, start));
	
			// l'URL
			String originale = matcher.group(1);
			logger.debug(THIS_CLASS + "original: " + originale);
			
			String remplacement;
			if (originale.contains("TogglePlus.gif"))
				remplacement = resourceHandler.handleTogglePlus();
			
			else if (originale.contains("ToggleMinus.gif"))
				remplacement = resourceHandler.handleToggleMinus();
			
			else if (originale.contains("ImageID"))
				remplacement = transformImageRequest(originale, streams);
			
			else if (originale.contains("ShowHideToggle"))
				remplacement = transformDrillDownRequest(originale);
			
			else
				remplacement = transformDrillThroughRequest(originale);
			
			SSRSWrapper.logger.debug(THIS_CLASS + "new     : " + remplacement);
			buffer.append(remplacement);
	
			// et on continue aprs l'URL
			lastEnd = end;
		}
		// on met la fin
		buffer.append(html.substring(lastEnd, html.length()));
		return buffer.toString();
	}

	/**
	 * cre l'url pour une image
	 * @param url l'url qui fait rfrence dans l'image dans le HTML produit par ReportingServices
	 * @param streams les ressources externes au rapport
	 * @return une URL pour charger l'image
	 * @throws GramException
	 */
	private String transformImageRequest(String url, Map<String,FileVO> streams)
			throws GramException {

		// on rcupre les paramtres
		String tokens[] = url.split("\\?")[1].split("&amp;");
	
		// get the resource ID
		String id = null;
		for (int i = 1; i < tokens.length; i++) {
			String subTokens[] = tokens[i].split("=");
			if (subTokens.length<2)
				continue;
			if (subTokens[0].equals("rs%3aImageID")) {
				id = subTokens[1];
				break;
			}
		}

		return resourceHandler.handleResource(session, streams.get(id));
	}

	/**
	 * lien vers un autre rapport
	 * @param url l'url d'origine
	 * @return l'url cre par le frontend
	 * @throws GramException
	 */
	public String transformDrillThroughRequest(String url)
			throws GramException {
		
		// on prend les paramtres de l'URL
		String tokens[] = url.split("\\?")[1].split("&amp;");

		// le chemin du rapport
		String rapport = tokens[0];
		
		// les paramtres du rapport
		Map<String, String> params = new HashMap<String, String>();
		for (int i = 1; i < tokens.length; i++) {
			// on ignore les paramtres de Reporting Services
			// TODO est-ce vraiment une bonne ide ?
			if (tokens[i].startsWith("rc%3a"))
				continue;
			if (tokens[i].startsWith("rs%3a"))
				continue;
			if (!tokens[i].contains("="))
				continue;
			String subTokens[] = tokens[i].split("=");
			if (subTokens.length < 2)
				continue;
			params.put(subTokens[0], subTokens[1]);
		}
	
		// on donne le tout  l'ActionHandler
		return actionHandler.handleDrillThrough(rapport, ReportFormat.HTML, params);
	}

	public String transformDrillDownRequest(String url)
			throws GramException {
	
		// on prend les paramtres
		String tokens[] = url.split("\\?")[1].split("&amp;");
	
		String itemID = null;
		for (int i = 1; i < tokens.length; i++) {
			String subTokens[] = tokens[i].split("=");
			if (subTokens.length < 2)
				continue;
			if (subTokens[0].equals("rs%3aShowHideToggle")) {
				itemID = subTokens[1];
				break;
			}
		}
		// TODO toggleId == null => exception
		return actionHandler.handleDrillDown(itemID);
	}
}