package acube.gram.business.impl;

import java.util.HashMap;

import javax.servlet.http.HttpServletRequest;

import acube.framework.vo.ErrorVO;
import acube.gram.backend.Backend;
import acube.gram.backend.ExecutionContext;
import acube.gram.business.Delegate;
import acube.gram.config.ReportConfig;
import acube.gram.exception.GramErrorCodes;
import acube.gram.exception.GramException;

/**
 * Implmentation de base d'un <i>delegate</i> GRaM. La liste des rapports et 
 * les droits sont dfinis dans un fichier XML. 
 * <br />
 * Les dfinitions de rapport sont dans un dossier de l'application (typiquement /WEB-INF/reports),
 * et le fichier de configuration doit lister pour chacun la liste des profils qui y ont accs.
 * Ceux qui ne sont pas dcrits dans le fichier de configuration sont inaccessibles.
 * <br />
 * Des rgles supplmentaires peuvent tre implmentes en surchargeant les mthodes protges 
 * de cette classe.
 * @author MAE
 *
 */
public class SimpleDelegate implements Delegate {

	private Backend backend;
	private HashMap<String, ReportConfig> reports = new HashMap<String,ReportConfig>();
	private boolean skipSecurityTests = true;
	
	public void setBackend(Backend backend) {
		this.backend = backend;
	}

	public void addReport(ReportConfig report) {

		this.reports.put(report.getPath(),report);
	}
	
	public void setSkipSecurityTests(boolean checkAuthorizations) {
		this.skipSecurityTests = checkAuthorizations;
	}

	/**
	 * cherche le rapport demand dans la liste des raports du fichier de configuration.
	 * S'il est trouv, vrifie que l'utilisateur a les droits ncessaires, charge le
	 * rapport et encapsule ventuellement le contexte d'excution obtenu dans un Delegate.
	 */
	public ExecutionContext loadReport(HttpServletRequest request, String path)
			throws GramException {

		// cherche le rapport
		ReportConfig report = reports.get(path);
		if (report==null) {
			ErrorVO error = new ErrorVO();
			error.setCdError(GramErrorCodes.INVALID_REPORT);
			error.setParamLabel(path);
		}

		// verifie que l'utilisateur a un des roles qui permettent de voir le rapport
		boolean allowed = skipSecurityTests || isUserAllowed(request, report);

		if (!allowed) {
			ErrorVO error = new ErrorVO();
			error.setCdError(GramErrorCodes.AUTHORIZATION_ERROR);
			error.setParamLabel(path);
			throw new GramException(error);
		}
		
		// initie le rapport
		ExecutionContext context = backend.loadReport(request, path);
		
		// encapsule le contexte d'excution si besoin est
		return embedExecutionContext(context);
	}

	/**
	 * Vrifie que l'utilisateur a bien le droit de consulter ce rapport.
	 * Par dfaut, les droits sont bass sur les profils de l'utilisateur :
	 * s'il a un des profils requis par le rapport, on autorise, sinon on refuse.
	 * 
	 * @param request la requte de l'utilisateur
	 * @param report la dfinition de rapport dans le fichier de proprits
	 * @return true si l'utilisateur est autoriser  consulter le rapport, false sinon
	 * @throws GramException
	 */
	protected boolean isUserAllowed(HttpServletRequest request, ReportConfig report)
			throws GramException {

		boolean allowed = false;
		for (String role : report.getRoles())
			if (request.isUserInRole(role)) {
				allowed = true;
				break;
			}
		return allowed;
	}
	
	/**
	 * Est appele aprs la cration du contexte d'excution par le <i>back-end</i>.
	 * Permet d'encapsuler le contexte dans un <i>delegate</i> pour les rgles spcifiques
	 * (en particulier les filtrage des valeurs de paramtre).
	 * La mthode par dfaut renvoie le contexte tel quel.
	 * 
	 * @param context le contexte brut
	 * @return le contexte qui sera utilis par le <i>front-end</i>
	 * @throws GramException
	 */
	protected ExecutionContext embedExecutionContext(ExecutionContext context)
			throws GramException {
		
		return context;
	}
}
