package acube.gram.util;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

import javax.servlet.http.HttpSession;

/**
 * Cache mmoire, typiquement en session.
 * Les objets sont stocks pour une certaine dure, et ont chacun un identifiant
 * qui permet de les rcuprer.
 * A chaque ajout/lecture, les objets prims sont vids du cache.
 *
 */
public class SimpleCache {

	/**
	 * La dure de validit par dfaut
	 */
	private static final long duration = 3600000; // one hour
	/**
	 * les objets stocks
	 */
	private HashMap<String, Resource> resources = new HashMap<String, Resource>();
	
	/**
	 * Rcupre un objet stock dans un cache session
	 * @param session la session
	 * @param sessionAttribute le nom de l'attribut de session qui contient le cache
	 * @param id le nom de l'objet
	 * @return l'objet stock
	 */
	public static Object get(HttpSession session, String sessionAttribute, String id) {
		
		return getStoreFromSession(session, sessionAttribute).get(id);
	}

	/**
	 * Ajoute un objet dans un cache session
	 * @param session la session
	 * @param sessionAttribute le nom de l'attribut de session qui contient le cache
	 * @param value l'objet  stocker
	 * @return l'identifiant de l'objet dans le cache
	 */
	public static String add(HttpSession session, String sessionAttribute, Object value) {
		
		return getStoreFromSession(session, sessionAttribute).add(value);
	}

	/**
	 * Rcupre le cache session, ou en cre un s'il n'existe pas
	 * @param session la session
	 * @param sessionAttribute le nom de l'attribut de session qui contient le cache
	 * @return le cache session
	 */
	private static SimpleCache getStoreFromSession(HttpSession session, String sessionAttribute) {

		SimpleCache store = (SimpleCache)session.getAttribute(sessionAttribute);
		
		if (store==null) {
			store = new SimpleCache();
			session.setAttribute(sessionAttribute, store);
		}
		return store;
	}
	
	/**
	 * Ajoute un objet au cache.
	 * @param resource l'objet  ajouter
	 * @return l'id donn  l'objet
	 */
	public synchronized String add(Object resource) {

		cleanup();
		String id = generateUniqueId();
		resources.put(id,new Resource(resource));
		return id;
	}

	/**
	 * Cherche un objet dans le cache.
	 * @param id l'identifiant de l'objet
	 * @return l'objet stock, ou null s'il n'est pas trouv (identifiant invalide ou prim)
	 */
	public Object get(String id) {
		
		cleanup();
		Resource resource = resources.get(id);
		if (resource==null)
			return null;
		return resource.content;
	}

	/**
	 * cre un nouvel identifiant. Principe : on cre un entier en "hashant" la date et l'heure,
	 * et on cherche  partir de l le premier entier encore non utilis.
	 * @return un identifiant encore non utilis
	 */
	private synchronized String generateUniqueId() {
		
		long time = new Date().getTime();
		do {
			String id = String.valueOf(new Long(time).hashCode());
			if (!resources.containsKey(id))
			{
				return id;
			}
			time++;
		} while (true);
	}
	
	/**
	 * retire les objets prims
	 */
	private void cleanup() {
		ArrayList<String> remove = new ArrayList<String>();
		for (String key : resources.keySet()) {
			Resource resource = resources.get(key);
			if (!resource.isValid())
				remove.add(key);
		}
		for (String key : remove)
			resources.remove(key);
	}
	
	/**
	 * Objet stocks dans un cache : objet+date d'expiration
	 */
	private static class Resource {
		
		/**
		 * l'objet stock
		 */
		private Object content;
		/**
		 * la "date de premption"
		 */
		private long expirationTime;
		
		/**
		 * utilise la dure standard 
		 * @param content l'objet ) stocker
		 */
		public Resource(Object content) {
			
			this.content = content;
			this.expirationTime = new Date().getTime() + duration;
		}
		
		/**
		 * utilse une dure spcifique
		 * @param content l'objet  staocker
		 * @param duration la dure de validit, en millisecondes
		 */
		public Resource(Object content, int duration) {
			
			this.content = content;
			this.expirationTime = new Date().getTime() + duration;
		}
		
		/**
		 * test de validit
		 * @return false si la date de premption est passe, true sinon
		 */
		boolean isValid() {
			return (new Date().getTime()<expirationTime);
		}	
	}
}
