package acube.gram2.business;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import acube.framework.exception.DAOException;
import acube.framework.exception.JDBCWrapperException;
import acube.framework.exception.TechnicalException;
import acube.gram.backend.Backend;
import acube.gram.exception.GramException;
import acube.gram2.integration.DAO;
import acube.gram2.vo.NodeVO;
import acube.gram2.vo.ReportVO;
import acube.gram2.vo.RoleVO;

public class DAODelegateImpl extends AbstractDelegateImpl {

	private String roleAdministrator = "admin";
	private String roleManager = "manager";
	private DAO dao;
	
	public DAODelegateImpl(Backend backend, DAO dao) {

		super(backend);
		this.dao = dao;
	}	
	
	@Override
	public boolean isAdministrator(HttpServletRequest request) throws GramException {

		return request.isUserInRole(roleAdministrator);
	}
	
	@Override
	public boolean isManager(HttpServletRequest request) throws GramException {

		return request.isUserInRole(roleManager);
	}

	public List<RoleVO> listRoles(HttpServletRequest request)
			throws TechnicalException, JDBCWrapperException, DAOException {
		
		return dao.listRoles();
	}

	public ReportVO getReportById(HttpServletRequest request, String path)
			throws TechnicalException, JDBCWrapperException, DAOException {
		
		return dao.getReportById(path);
	}

	public List<ReportVO> listReports(HttpServletRequest request)
			throws TechnicalException, JDBCWrapperException, DAOException {
		
		// list the reports
		List<ReportVO> reports = dao.listReports();
		// add roles
		for (ReportVO report : reports) {
			for (RoleVO role : dao.listRolesForReport(report.getPath()))
				report.addRole(role.getId());
		}
		return reports;
	}

	public void createReport(HttpServletRequest request, String path, String name, List<String> roles)
		throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		dao.addReport(path, name);
		for (String role : roles)
			dao.addRoleToReport(path, role);
	}

	public void updateReport(HttpServletRequest request, String path, String name, List<String> roles)
			throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		// update the name
		dao.updateReport(path, name);

		// update the role list
		List<String> oldRoles = new ArrayList<String>();
		for (RoleVO role : dao.listRolesForReport(path))
			oldRoles.add(role.getId());

		// add the new roles
		for (String role : roles)
			if (!oldRoles.remove(role))
				dao.addRoleToReport(path, role);

		// roles that are left are no longer needed : remove them
		for (String role : oldRoles)
			dao.removeRoleFromReport(path, role);
	}

	public void deleteReport(HttpServletRequest request, String path)
			throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		dao.removeAllRolesFromReport(path);
		dao.removeReportFromAllNodes(path);
		dao.deleteReport(path);
	}
	
	
	
	
	
	public List<NodeVO> listNodes(HttpServletRequest request)
			throws TechnicalException, JDBCWrapperException, DAOException {
			
		HashSet<String> visibleReports = new HashSet<String>();
		// all reports are visible for managers and administrators
		if (isAdministrator(request) || isManager(request))
			for (ReportVO report : dao.listReports())
				visibleReports.add(report.getPath());
		
		// for the other users, only those for their roles
		else
			for (ReportVO report : dao.listReports())
				for (RoleVO role : dao.listRoles())
					if (request.isUserInRole(role.getId())) {
						visibleReports.add(report.getPath());
						break;
					}

		Map<Integer,NodeVO> nodes = new HashMap<Integer,NodeVO>();
		List<NodeVO> roots = new ArrayList<NodeVO>();
		
		for (NodeVO node : dao.listNodes()) {
			nodes.put(node.getId(), node);
			if (node.getParent()==null)
				roots.add(node);
		}
		
		for (NodeVO node : nodes.values()) {
			// link the node to its parent
			Integer parent = node.getParent();
			if (parent != null)
				nodes.get(parent).addChildNode(node);
			// adds the children reports
			for (ReportVO report : dao.listReportsForNode(node.getId()))
				if (visibleReports.contains(report.getPath()))
					node.addChildReport(report);
		}
		
		return filterEmptyNodes(roots);
	}
	
	public List<NodeVO> filterEmptyNodes(List<NodeVO> nodes) {
		
		List<NodeVO> result = new ArrayList<NodeVO>();
		for (NodeVO node : nodes) {
			node.setChildNodes(filterEmptyNodes(node.getChildNodes()));
			if (!node.getChildNodes().isEmpty() || !node.getChildReports().isEmpty())
				result.add(node);
		}
		return result;
	}

	public void createNode(HttpServletRequest request, Integer parent, String name, List<String> reports)
		throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		int id = dao.addNode(parent, name);
		for (String report : reports)
			dao.addReportToNode(id, report);
	}

	public void updateNode(HttpServletRequest request, int id, String name, List<String> reports)
		throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		dao.updateNode(id, name);

		// update the report list
		List<String> oldReports = new ArrayList<String>();
		for (ReportVO role : dao.listReportsForNode(id))
			oldReports.add(role.getPath());

		// add the new reports
		for (String report : reports)
			if (!oldReports.remove(report))
				dao.addReportToNode(id, report);

		// reports that are left are no longer needed : remove them
		for (String report : oldReports)
			dao.removeReportFromNode(id, report);
	}

	public void deleteNode(HttpServletRequest request, int id)
		throws TechnicalException, JDBCWrapperException, DAOException {

		assertManager(request);
		dao.removeAllReportsFromNode(id);
		dao.deleteNode(id);
	}
}
